package cafe;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

/**
 * A cafe implemented by a circular queue of Customer objects.
 */
public class CafeQueue
{
    private Customer customers[];  // array of Customer objects
    private int head;              // queue head index (customers served here)
    private int tail;              // queue tail index (customers entered here)
    private int customersCount;    // count of customers in the cafe 
    private boolean makingBurger;  // true when making a burger
    
    // Queue lock and its conditions.
    private Lock qLock = new ReentrantLock();
    private Condition spaceAvailable  = qLock.newCondition();
    private Condition patronAvailable = qLock.newCondition();
    private Condition burgerDone      = qLock.newCondition();
    
    long start;  // simulation start time

    /**
     * Construct an empty queue.
     * @param capacity the maximum capacity of the queue.
     */
    public CafeQueue(int capacity) 
    {
        customers = new Customer[capacity];
        head = 0;
        tail = 0;
        customersCount = 0;
        makingBurger = false;
        start = System.currentTimeMillis();
    }

    /**
     * Add a Customer object at the tail of the queue.
     * @param customer the object to be added. 
     * @throws InterruptedException if interrupted while waiting.
     */
    public void add(Customer customer) throws InterruptedException
    {
        qLock.lock();
       
        try {
            // While making a burger, wait for the burger to be done.
            while (makingBurger) burgerDone.await();
            
            // While full, wait for space to become available.
            while (isFull() || makingBurger) spaceAvailable.await();
            
            makingBurger = customer.getType() == CustomerType.BURGER;
            
            // Add the customer to the tail of the queue.
            customers[tail] = customer;
            tail++;
            customersCount++;
            if (tail == customers.length) tail = 0;
            
            print(customer);
            
            // There's now at least one customer in the cafe.
            patronAvailable.signalAll();
        }
        finally {
            qLock.unlock();
        }
    }
    
    /**
     * Remove (serve) a Customer object at the head of the queue.
     * @return the object that was removed. 
     * @throws InterruptedException if interrupted while waiting.
     */
    public Customer remove() throws InterruptedException
    {
        qLock.lock();
        
        try {
            // If the cafe is empty, wait until there's a customer.
            while (isEmpty()) patronAvailable.await();
            
            // Remove (serve) the customer at the head of the queue.
            Customer customer = customers[head];
            head++;
            customersCount--;
            if (head == customers.length) head = 0;
            
            // If it was a burger customer, signal that the burger is done.
            if (customer.getType() == CustomerType.BURGER) {
                makingBurger = false;
                burgerDone.signalAll();
            }
            
            // Print the customer with the sequence number negated.
            customer.negateSequence();
            print(customer);
            customer.negateSequence();
            
            // Now there's at least one space available in the cafe.
            spaceAvailable.signalAll();
            
            return customer;
        }
        finally {
            qLock.unlock();
        }
    }
    
    /**
     * @return the Customer object at the head of the queue. 
     * @throws InterruptedException if interrupted while waiting.
     */
    public Customer peek() throws InterruptedException
    {
        qLock.lock();
        
        try {
            // If the cafe is empty, wait until there's a customer.
            while (isEmpty()) patronAvailable.await();
            
            // Return the customer at the head of the queue.
            return customers[head];
        }
        finally {
            qLock.unlock();
        }
    }
    
    /**
     * @return true if the cafe is full.
     */
    public boolean isFull() 
    {
        return customersCount == customers.length;
    }

    /**
     * @return true if the cafe is empty.
     */
    public boolean isEmpty() 
    {
        return customersCount == 0;
    }
    
    /**
     * Print a customer that entered or left the cafe.
     * Include the timestamp and the current number of customers in the cafe.
     * @param customer the Customer object.
     */
    private void print(Customer customer)
    {
        long elapsed = System.currentTimeMillis() - start;
        int minutes = (int) elapsed/1000;
        double seconds = 60*((elapsed%1000)/1000.0);
        
        System.out.printf("%2d:%05.2f(%d) %s\n", 
                          minutes, seconds, customersCount, customer);
    }
}
